open Ast
open Sast

let rec c_var_type (typ: Ast.var_type) = match typ with
    Int -> "int"
  | Float -> "double"
  | Bool -> "bool"
  | Char -> "char"
  | Str -> "string"
  | Graph -> "Graph"
  | Node -> "Node"
  | List(vtype) -> "List<" ^ c_var_type vtype ^ ">" 
  | Dict(v1, v2) -> "Dict<" ^ c_var_type v1 ^ "," ^ c_var_type v2 ^ ">"
  | Void -> "void"


and c_var_decl (v: Sast.s_var_decl) = 
  let str = c_var_type v.s_vtype ^ " " ^ v.s_vname in
  if v.s_vinit.e = Noexpr then
    str
  else
    str ^ " = " ^ c_expr v.s_vinit 

and c_expr (e: Sast.s_expr) = c_ast_expr e.e 

and c_ast_expr (e: Ast.expr) = match e with
    IntLit(lit)   -> string_of_int lit
  | FloatLit(lit) -> string_of_float lit
  | BoolLit(lit)  -> string_of_bool lit
  (* TODO *)
  | CharLit(lit)  -> "'" ^ Char.escaped lit ^ "'"
  | StrLit(lit)   -> "\"" ^ lit ^ "\""
  | Id(id)        -> id
  | Assign(e1, asnop, e2) ->  "(" ^ c_ast_expr e1  ^ string_of_asnop asnop ^ c_ast_expr e2 ^ ")"
  | Not(expr)     -> "(!" ^ c_ast_expr expr ^ ")"
  | Binop(expr1,op,expr2) -> "(" ^ c_ast_expr expr1 ^ " " ^ string_of_binop op ^ " "
                             ^ c_ast_expr expr2 ^ ")"
  | Call(fname,param)  -> fname ^ "(" ^ 
                          (String.concat ", " (List.map c_ast_expr param)) ^ ")" 
  (* TODO *)
  | Value(expr1,expr2) -> "(" ^ c_ast_expr expr1 ^ "[" ^ c_ast_expr expr2 ^ "])"
  | ListLit(elements) -> "ListLit( [" ^ String.concat "; " (List.map c_ast_expr elements) ^ "] )" 
  | SetLit(elements) -> "SetLit( [" ^ String.concat ", " (List.map c_ast_expr elements) ^ "] )"
  | DictLit(kvpairs) -> "DictLit( [ TODO ] )"
  | GraphLit(edges) -> 
      "({Graph _tmp; " ^ (String.concat "; " 
      (List.map (fun edge -> "_tmp.insertEdge(" ^ c_ast_expr edge.src ^ ", " 
                             ^ c_ast_expr edge.label ^ ", " ^ c_ast_expr
                             edge.dest ^ ")" ) edges)) ^ "; _tmp;})"   
  | Null -> "null"
  | Noexpr -> ""


and c_stmt (s: Sast.s_stmt) = match s with 
    S_Block(_,sl) -> "{\n" ^ (String.concat ";\n" (List.map c_stmt sl)) ^ ";\n}\n"
  | S_Variable(var) -> c_var_decl var
  | S_Expr(expr) -> "(" ^ (c_expr expr) ^ ")"
  | S_Return(expr) -> "return " ^ (c_expr expr)
  | S_If(expr, s1, s2) -> "if (" ^ c_expr expr ^ ")\n" 
                          ^ c_stmt s1 ^ ";\nelse" ^  c_stmt s2
  | S_For(init, test, after, stmt) -> "for (" ^ c_expr init ^ "; " ^ c_expr test
                                      ^ "; " ^ c_expr after ^ ") " ^ c_stmt stmt
  | S_Foreach(e1, e2, stmt) -> "for (" ^ c_expr e1 ^ " : " ^ c_expr e2 ^ ") " ^ c_stmt stmt
  | S_While(test, stmt) -> "while ( " ^ c_expr test ^ " )" ^ c_stmt stmt 
  | S_Continue          -> "continue"
  | S_Break             -> "break"



let c_func_decl (f: Sast.s_func_decl) =
  let fname = f.s_fname in 
  c_var_type f.s_rtype ^ " " ^ fname ^ "("
  ^ (String.concat ", " (List.map (fun f -> (c_var_type (fst f)) ^ " " ^ snd f) f.s_formals))
  ^ ") {\n" ^ (String.concat ";\n" (List.map c_stmt f.s_body)) ^ ";\n}"


let c_program program = 
  let global_vars = (List.map c_var_decl program.s_var_decls) in
  let func_decls = (List.map c_func_decl program.s_func_decls) in
  "#include <iostream>\n#include <string>\n#include \"test.h\"\n"
  ^ "\nusing namespace std;\n\n"
  ^ (String.concat ";\n" global_vars) ^ "\n"
  ^ (String.concat "\n" func_decls)
  ^ "\nint main() { __main(); return 0; }"  




