#ifndef __GRAPH_H_
#define __GRAPH_H_


#include <iostream>
#include <string>
#include <vector>
#include <unordered_map>
#include <map>
#include <algorithm>

using namespace std; 

class Graph;

//defines a Node type
class Node {
public:
  //initializes a node with the given string as the node's name
  //and the graph where this node belongs to
  Node(string s,Graph *g) 
  {
    this -> name = s;
    this -> g = g;
    neighborSize=0;
  }

  //returns the size of this neighbors which is the sum of 
  //outNeighbors and incomingNeighbors
  int getNeighborSize()
  {
    return neighborSize;
  }

  //returns the neighbors that are connected by the given
  //relationship label where this node is the source
  vector<Node*> getOutNeighbors(string label){
    if(outNeighbors.find(label)==outNeighbors.end())
    {
      return vector<Node*>();
    }
    return outNeighbors[label];
  }

  //returns all neighbor nodes where this node is the source
  vector<Node*> getOutNeighbors(){
    vector<Node*> result;
    for(auto it=outNeighbors.begin();it!=outNeighbors.end();it++){
      vector<Node*> curr = it->second;
      for(int i=0;i<curr.size();++i)
      {
        if(find(result.begin(),result.end(),curr[i])==result.end())
        {
          result.push_back(curr[i]);
        }
      }
    }
    return result;
  }

  unordered_map<string,vector<Node*> >& getOutNeighborsMap()
  {
    return outNeighbors; 
  }

  //returns the neighbors that are connected by the given
  //relationship label where this node is the target
  vector<Node*> getInNeighbors(string label) 
  {
    if(inNeighbors.find(label)==inNeighbors.end())
    {
      return vector<Node*>();
    }
    return inNeighbors[label];
  }

  unordered_map<string,vector<Node*> >& getInNeighborsMap()
  {
    return inNeighbors; 
  }

  //returns all neighbor nodes where this node is the source
  vector<Node*> getInNeighbors(){
    vector<Node*> result;
    for(auto it=inNeighbors.begin();it!=inNeighbors.end();it++){
      vector<Node*> curr = it->second;
      for(int i=0;i<curr.size();++i)
      {
        if(find(result.begin(),result.end(),curr[i])==result.end())
        {
          result.push_back(curr[i]);
        }
      }
    }
    return result;
  }

  //addes a neighbor to this node
  void addNeighbor(Node &node,string label){
    addToOutNeighbors(node,label);
    node.addToInNeighbors(*this,label);
  }

  void addToOutNeighbors(Node &node, string label)
  {
    if(outNeighbors.find(label)==outNeighbors.end()){
      vector<Node*> curr;
      curr.push_back(&node);
      outNeighbors.emplace(label,curr);
    }
    else{
      outNeighbors[label].push_back(&node);
    }
    neighborSize+=1;
  }

  void addToInNeighbors(Node &node, string label)
  {
    if(inNeighbors.find(label)==inNeighbors.end()){
      vector<Node*> curr;
      curr.push_back(&node);
      inNeighbors.emplace(label,curr);
    }
    else{
      inNeighbors[label].push_back(&node);
    }
    neighborSize+=1;
  }

  //removes a neighbor from this node
  void removeNeighbor(Node& node)
  {
    removeOutNeighbor(node);
    node.removeInNeighbor(*this);
  }

  void removeOutNeighbor(Node& node)
  {
    for(auto it=outNeighbors.begin();it!=outNeighbors.end();++it)
    {
      vector<Node*>& curr = it->second;
      auto iter = find(curr.begin(),curr.end(),&node);
      if(iter!=curr.end())
      {
        curr.erase(iter);
      }
    }
    neighborSize-=1;
  }

  void removeInNeighbor(Node& node)
  {
    cout << "removeInNeighbor : " << node.getName() << endl;
    for(auto it=inNeighbors.begin();it!=inNeighbors.end();++it)
    {
      vector<Node*>& curr = it->second;
      cout<< curr.size() << endl;
      auto iter = find(curr.begin(),curr.end(),&node);
      if(iter!=curr.end())
      {
        cout << "removed" << endl;
        curr.erase(iter);
      }
      cout << curr.size() <<endl;
    }
    neighborSize-=1;
  }

  //adds attribute to this node
  void addAttribute(string key, string value)
  {
    attributes[key]=value;
  }

  //removes a attribute from this node
  void removeAttribute(string key)
  {
    if(attributes.find(key)!=attributes.end())
    {
      attributes.erase(key);
    }
  }

  //returns the name of this node
  string getName() const
  {
    return name;
  }

  //returns the graph where this node belongs to
  const Graph* getGraph() const
  {
    return g;
  }

  bool operator==(const Node& other) const
  {
    return g==other.getGraph() && name == other.getName();
  }

  unordered_map<string,string>& getAttributes()  
  {
    return attributes;
  }

  friend ostream &operator<<(ostream &os, Node &node);
  friend bool _hasAttribute(const Node *n,const string &key);
  friend unordered_map<string,string>& _getAttributes(Node *n);  
  friend vector<Node*> getOutNeighbors(Node *n,string label);
  friend vector<Node*> getInNeighbors(Node *n,string label);

private:
string name;
const Graph *g;
 unordered_map<string,vector<Node*> > outNeighbors;
 unordered_map<string,vector<Node*> > inNeighbors;
 unordered_map<string,string> attributes;
 int neighborSize;
};

bool _hasAttribute(Node *n,const string &key)
{
  unordered_map<string,string>& attributes = n->getAttributes();
  return attributes.find(key) != attributes.end();
}

unordered_map<string,string>& _getAttributes(Node *n)
{
  return n->getAttributes();
}

vector<Node*> getOutNeighbors(Node *n,string label)
{
  return n->getOutNeighbors(label);
}

vector<Node*> getInNeighbors(Node *n,string label)
{
  return n->getInNeighbors(label);
}

//prints the attributes, outgoing neighbors, and incoming neighbors of this node
ostream &operator<<(ostream &os, Node &node) {
  os << "Node : " + node.getName() + "\n";
  os << "-----------------------------\n";
  unordered_map<string,vector<Node*> > &oNeighbors = node.getOutNeighborsMap();
  for(auto it = oNeighbors.begin();it!=oNeighbors.end();it++)
  {
    os << "Outgoing Relation: " + it->first + "\n";
    vector<Node*>& neighbors = it->second;
    for(int i=0;i<neighbors.size();i++)
    {
      os<< neighbors[i]->getName() + "\n";
    }
  }
  unordered_map<string,vector<Node*> > &iNeighbors = node.getInNeighborsMap();
  os << "-----------------------------\n";
  for(auto it = iNeighbors.begin();it!=iNeighbors.end();it++)
  {
    os << "Incoming Relation: " + it->first + "\n";
    vector<Node*>& neighbors = it->second;
    for(int i=0;i<neighbors.size();i++)
    {
      os<< neighbors[i]->getName() + "\n";
    }
  }
  unordered_map<string,string>& attrs = node.getAttributes();
  os << "------------------------------\n";
  os << "Attributes:\n ";
  for(auto it = attrs.begin();it!=attrs.end();it++)
  {
    os<< it->first + " : " + it->second + "\n";
  }
  return os;
}

namespace std
{
  template<>
  struct hash<Node>
  {
    size_t operator()(const Node& k) const
    {
      return hash<string>()(k.getName());
    }
  };
}

//defines a graph type
class Graph {
public:
  Graph(){}

  ~Graph() {
    for(auto it = nodes.begin();it!=nodes.end();it++)
    {
      delete(it->second);
    }
  }
 
  static Graph getGraph() {return Graph(); }
  
  //inserts an edge to the graph by creating nodes with given names
  //as source and dest
  void insertEdge(string source, string label, string dest) {
    if (nodes.find(source) == nodes.end()) {
      // new Node ()
      // --> create NodeRef -> vector
      Node *sNode  = new Node(source,this); 
      nodes[source]=sNode;
    }
    if (nodes.find(dest) == nodes.end()) {
      Node *tNode = new Node(dest,this);
      nodes[dest] = tNode;
    }

    Node *sourceNode = nodes[source];
    Node *targetNode = nodes[dest];
    sourceNode->addNeighbor(*targetNode,label);
    edges[make_pair(source, dest)].push_back(label);
  }

  Node& getNode(const string& name){
    return *nodes[name];
  }  

  unordered_map<string,Node*>& getAllNodes()
  {
    return nodes;
  }

  //returns all the relations between n1 and n2
  vector<string> getLabels(Node& n1,Node& n2)
  {
    string node1 = n1.getName();
    string node2 = n2.getName();
    pair<string,string> p1(node1,node2);
    for(auto it = edges.begin();it!=edges.end();it++)
    {
      pair<string,string> key = it -> first;
      if(key==p1)
      {
        return it->second;
      }
    }
    return vector<string>();
  }

  map<pair<string,string>,vector<string> >& getEdges()
  {
    return edges;
  }

  //merges this graph with the given graph g
  void mergeGraph(Graph &g)
  {
    unordered_map<string,Node*>& newNodes = g.getAllNodes();
    for(auto it=newNodes.begin();it!=newNodes.end();it++)
    {
      if(nodes.find(it->first)==nodes.end())
      {
        nodes[it->first]=it->second;
      }
    }
    newNodes.clear();

    map<pair<string,string>,vector<string>>& newEdges = (map<pair<string,string>,vector<string>>&)g.getEdges();
    for(auto it=newEdges.begin();it!=newEdges.end();it++)
    {
      pair<string,string>& curr =(pair<string,string>&) it->first;
      if(edges.find(curr) == edges.end())
      {
        edges.emplace(curr,it->second);
        vector<string>& newLabels = it->second;
        for(auto it = newLabels.begin();it!=newLabels.end();it++)
        {
            getNode(curr.first).addNeighbor(getNode(curr.second),*it);
        }
      }
      else
      {
        vector<string>& newLabels = it->second;
        vector<string>& currLabels = edges[curr];
        for(auto it = newLabels.begin();it!=newLabels.end();it++)
        {
          if(find(currLabels.begin(),currLabels.end(),*it)==currLabels.end())
          {
            currLabels.push_back(*it);
            getNode(curr.first).addNeighbor(getNode(curr.second),*it);
          }
        }
      }
    }
  }

  //deletes the given graph g from this graph
  void deleteGraph(Graph& g)
  {
    map<pair<string,string>,vector<string> >& subgraphEdges =(map<pair<string,string>,vector<string> >&)g.getEdges();
    for(auto it = subgraphEdges.begin();it!=subgraphEdges.end();it++)
    {
      pair<string,string>& curr = (pair<string,string>&)it->first;
      if(edges.find(curr)!=edges.end())
      {
        edges.erase(curr);
        Node *source = &getNode(curr.first);
        Node *target = &getNode(curr.second);
        source->removeNeighbor(getNode(curr.second));
        if(source->getNeighborSize()==0)
        {
          nodes.erase(curr.first);
          delete(source);
        }
        if(target->getNeighborSize()==0)
        {
          nodes.erase(curr.second);
          delete(target);
        }
      }
    }
  }

  friend ostream &operator<<(ostream &os, Graph const &g);
  friend Node* _getNode(Graph& g,const string& name);
  friend Graph& _plusequal(Graph& g,Graph& h);
  friend Graph& _plus(Graph &g, Graph& h);
  friend void _insertEdge(Graph &g,string source,string label,string dest);
  friend vector<Node*> getNodes(Graph &g);
  friend vector<string> getLabels(Graph &g);
  friend void _minusEqualGraph(Graph& g, Graph &h);
  friend Graph& _minusGraph(Graph& g,Graph &h);

private:
  map<pair<string, string>, vector<string> > edges;
  unordered_map<string,Node*> nodes;
};

void _insertEdge(Graph& g,string source,string label,string dest)
{
  g.insertEdge(source,label,dest);
}

void _minusEqualGraph(Graph &g, Graph &h)
{
  g.deleteGraph(h);
}

Graph& _minusGraph(Graph &g,Graph &h)
{
  Graph *result = new Graph();
  result->deleteGraph(h);
  return *result;
}

vector<Node*> getNodes(Graph &g)
{
  vector<Node*> result;
  for(auto it=g.nodes.begin();it!=g.nodes.end();it++)
  {
    result.push_back(it->second);
  }
  return result;
}

vector<string> getLabels(Graph &g)
{
  vector<string> result;
  for(auto it=g.edges.begin();it!=g.edges.end();it++)
  {
    vector<string>& curr = it->second;
    for(int i=0;i<curr.size();i++)
    {
      result.push_back(curr[i]);
    }
  }
  return result;
}

Node* _getNode(Graph& g,const string& name)
{
  return &(g.getNode(name));
}

Graph& _plusequal(Graph &g, Graph& h)
{
  g.mergeGraph(h);
  return g;
}

Graph& _plus(Graph &g, Graph &h)
{
  Graph *result = new Graph();
  map<pair<string, string>, vector<string> >& edges = g.getEdges();
  unordered_map<string,Node*>& nodes = g.getAllNodes();
  for(auto it = edges.begin();it!=edges.end();it++)
  {
    vector<string>& labels = it->second;
    for(int i=0;i<labels.size();i++)
    {
      result->insertEdge(it->first.first,labels[i],it->first.second);
    }
  }
  result->mergeGraph(h);
  return *result;
}

Graph& _newGraph()
{
  return *(new Graph());
}

ostream &operator<<(ostream &os, const Graph &g) {
  os << "Graph {\n";
  //for (auto it = g.getEdges().begin(); it != g.getEdges().end(); it++) {
  const map<pair<string, string>, vector<string>> & es = g.edges;
  for (auto it = es.begin(); it != es.end(); it++) {
    for (int i = 0; i < it->second.size(); i++) {
      os << "Node(" << it->first.first << ") --" << it->second[i];
      os << "--> Node(" << it->first.second << ")\n";  
    }
  }
  os << "}";
  return os;
}


struct edge {
  string source, label, dest;
  edge(const string &s, const string &l, const string &d): source(s), 
  label(l), dest(d) {}
};

edge* _createEdge(string src, string label, string dest) {
  return new edge(src, label, dest);
}

Graph& _createGraph(int num, ...) {
  va_list edges;
  va_start(edges, num);
  Graph *g = new Graph();
  for (int i = 0; i < num; ++i) {
    edge *s = va_arg(edges, edge*);
    g->insertEdge(s->source, s->label, s->dest);
    delete s;
  }
  va_end(edges);
  return *g;
}

/* Built-in functions */
void print(int t) {
  cout << t;
}

void print(double t) {
  cout << t;
}

void print(char t) {
  cout << t;
}

void print(const string &t) {
  cout << t;
}

void print(const Graph &t) {
  cout << t;
}

void print(const Node *t) {
  cout << t->getName();
}

string getName(const Node *t) {
  return t->getName();
}




#endif
