/* 
 * KGL parser
 */


%{
  open Ast;; 
%}


/* Punctuation Tokens */
%token LPAREN RPAREN LBRACE RBRACE LBRACK RBRACK
%token SEMI COLON COMMA
%token LLIST RLIST LSET RSET LGRAPH RGRAPH LEDGE REDGE
%token EOF

/* Operator Tokens*/
%token PLUS MINUS MULTI DIVIDE MOD 
%token AND OR NOT
%token EQUAL NEQ GEQ LEQ GT LT
%token ASSIGN ADD SUB

/* Keywords */
%token FOR WHILE CONTINUE BREAK
%token IF ELSE
%token FUNC RETURN
%token INT FLOAT CHAR BOOL STR GRAPH NODE LIST DICT 
%token IN
%token NULL VOID

/* Literals */
%token <int> INT_LIT
%token <float> FLOAT_LIT
%token <string> ID STR_LIT
%token <char> CHAR_LIT
%token <bool> BOOL_LIT


/* Precedence, Associativity */
%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN ADD SUB
%left OR
%left AND
%left IN
%left EQUAL NEQ
%left GT GEQ LT LEQ
%left PLUS MINUS
%left MULTI DIVIDE MOD
%right NOT

%start test
%type <Ast.program> test
%type <Ast.program> program
%type <Ast.expr> expr
%type <Ast.var_type> var_type
%type <Ast.kv> kv
%type <Ast.edge> edge

%% 

test:
  program EOF   { $1 }


/* Program: var_decl_init; func_decl; ...  */
program:
  | /* nothing */               { [], [] }
  | program var_decl_init SEMI  { ($2 :: fst $1), snd $1 }
  | program func_decl           { fst $1, ($2 :: snd $1) } 


/* Function Declaration: func <ret_type> <id>(arg1, arg2, ...) { statment } */ 
func_decl:
  FUNC var_type ID LPAREN formals_opt RPAREN 
  LBRACE stmts_list RBRACE
  {
    {
      rtype = $2;
      fname = $3;
      formals = $5;
      body = List.rev $8; 
    }
  }


/* Variable Declaration: <type> <id>; <type> <id> = <initializer>  */ 
var_decl_init:
    var_decl                    { $1 }
  | var_type ID ASSIGN expr     { {v_type = $1; v_name = $2; v_init =  $4;} }

var_decl:
    var_type ID                 {  
                                  if ($1 == Void) then 
                                    raise Variable_of_void
                                  else 
                                    {v_type = $1; v_name = $2; v_init = Noexpr}  
                                }


/* Types */
var_type:
  | INT                                 { Int }   
  | FLOAT                               { Float }
  | CHAR                                { Char }
  | BOOL                                { Bool }
  | STR                                 { Str }
  | GRAPH                               { Graph }
  | NODE                                { Node }
  | LIST LT var_type GT                 { List($3) }
  | DICT LT var_type COMMA var_type GT  { Dict($3, $5) }
  | VOID                                { Void }


/* Statements */
stmts_list:
    /* nothing */               { [] }
  | stmts_list stmt             { $2 :: $1 }


stmt:
    LBRACE stmts_list RBRACE    { Block(List.rev $2) } 
  | var_decl_init SEMI          { Variable($1) } 
  | expr SEMI                   { Expr($1) }
  | RETURN expr_opt SEMI            { Return($2) } 
  | IF LPAREN expr_opt RPAREN stmt %prec NOELSE  { If($3, $5, Expr(Noexpr)) } 
  | IF LPAREN expr_opt RPAREN stmt ELSE stmt     { If($3, $5, $7) } 
  | FOR LPAREN expr_opt SEMI expr_opt SEMI expr_opt RPAREN stmt
                                             { For($3, $5, $7, $9) }  
  | FOR LPAREN expr RPAREN stmt              { Foreach($3, $5) }
  | WHILE LPAREN expr_opt RPAREN stmt        { While($3, $5) }   
  | CONTINUE SEMI                            { Continue } 
  | BREAK SEMI                               { Break } 


/* Formal Arguments */
formals_opt:
    /* nothing */               { [] }
  | formals_list                { List.rev $1 } 

formals_list:
    var_decl                    { [$1] }
  | formals_list COMMA var_decl { $3 :: $1 }


/* Actual Parameters */ 
actuals_opt:
  /* nothing */   { [ ] }
  | actuals_list  { List.rev $1}

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }


/* Expressions */
expr_opt:
  /* nothing */                 { Noexpr }
  | expr                        { $1 }  

expr:
    postfix_expr                { $1 }
  | NOT expr                    { Not($2) } 
  | expr MULTI expr             { Binop($1, Multi, $3) } 
  | expr DIVIDE expr            { Binop($1, Divide, $3) }
  | expr MOD expr               { Binop($1, Mod, $3) }
  | expr PLUS expr              { Binop($1, Plus, $3) }
  | expr MINUS expr             { Binop($1, Minus, $3) }
  | expr EQUAL expr             { Binop($1, Equal, $3) }
  | expr NEQ expr               { Binop($1, Nequal, $3) }
  | expr GT expr                { Binop($1, Greater, $3) }
  | expr GEQ expr               { Binop($1, Geq, $3) }
  | expr LT expr                { Binop($1, Less, $3) }
  | expr LEQ expr               { Binop($1, Leq, $3) }
  | expr IN expr                { Binop($1, In, $3) }
  | expr AND expr               { Binop($1, And, $3) }
  | expr OR expr                { Binop($1, Or, $3) }
  | expr ASSIGN expr            { Assign($1, Asn, $3) } 
  | expr ADD expr               { Assign($1, Add, $3) }
  | expr SUB expr               { Assign($1, Sub, $3) }
  | NULL                        { Null } 


/* Postfix Expressions */
postfix_expr:
    primary_expr                        { $1 }
  | ID LPAREN actuals_opt RPAREN        { Call($1, $3) } 
  | postfix_expr LBRACK expr RBRACK     { Value($1, $3) } 


/* Primary Expresssions */
primary_expr:
    literal                     { $1 }
  | ID                          { Id($1) }
  | LPAREN expr RPAREN          { $2 }  


/* Key/Value Pairs:  k1:v1, k2:v2, ... */
kvs_opt:
  /* nothing */                 { [ ] }
  | kvs_list                    { List.rev $1 } 

kvs_list:
  | kv                          { [$1] }
  | kvs_list COMMA kv           { $3 :: $1 } 

kv:
  | expr COLON expr             { {key = $1; value = $3} } 


/* Edges: src--(label)-->dest; ... */
edges_opt:
  /* nothing */                 { [ ] }
  | edges_list                  { List.rev $1 }

edges_list:
  | edge                        { [$1] }
  | edges_list SEMI edge        { $3 :: $1 }

edge:
  | expr LEDGE expr REDGE expr  { {src = $1; dest = $5; label = $3} }  


/* Literals */
literal:  
    INT_LIT                     { IntLit($1) }  
  | FLOAT_LIT                   { FloatLit($1) }
  | BOOL_LIT                    { BoolLit($1) }
  | CHAR_LIT                    { CharLit($1) }
  | STR_LIT                     { StrLit($1)  }
  | LGRAPH edges_opt RGRAPH     { GraphLit($2) }
  | LLIST actuals_opt RLIST     { ListLit($2)  }
  | LSET kvs_opt RSET           { DictLit($2) }

 

