(* 
 * KGL Semantic analysis: type checking and generation of SAST  
 *)

open Ast
open Sast
exception Error of string


let rec find_variable (scope: Sast.symbol_table) name =
  try
    List.find (fun (var, _) -> var = name) scope.variables
  with Not_found ->
    match scope.parent with
      Some(parent) -> find_variable parent name 
    | _ -> raise Not_found


let find_func (fname: string) (opts: var_type list)
              (funcs: (var_type * string * var_type list) list) : (var_type)= 
  let fl = (List.filter (fun (_, f, params) -> f = fname && params = opts) funcs) in
  if (List.length fl) = 0 then
    raise Not_found
  else
    let (ret_type, _, _) = List.hd fl in ret_type


let rec check_expr (e: Ast.expr) (env: environment) = match e with
    IntLit(lit) -> S_IntLit(lit), Int
  | FloatLit(lit) -> S_FloatLit(lit), Float
  | BoolLit(lit) -> S_BoolLit(lit),  Bool
  | CharLit(lit) -> S_CharLit(lit), Char
  | StrLit(lit) -> S_StrLit(lit), Str
  | Id(v) -> let (_, typ) = try
               find_variable env.scope v
             with Not_found ->
               raise (Error("Undeclared identifier " ^ v));
             in S_Id(v, typ), typ
  | Assign(e1, asnop, e2) -> check_assign e1 asnop e2 env 
  | Not(e) -> 
      let (e1, t1) = check_expr e env in
        if not (check_bool_valued t1) then 
          raise (Error("Type mismatch is operator ! " ^ string_of_var_type t1));
        S_Not(e1, t1), Bool
  | Binop(e1, op, e2) when op <> In -> 
      let (e1, t1) = check_expr e1 env and (e2, t2) = check_expr e2 env in
      let typ = check_binop_type t1 op t2 env in
      S_Binop(e1, op, e2, typ), typ
  | Binop(e1, In, e2) ->
      let (e1, t1) = check_expr e1 env and (e2, t2) = check_expr e2 env in
      if not (check_in_expr t1 t2) then
        raise (Error("Type mismatch in operator " ^ Ast.string_of_binop In));
      S_In(e1, e2, t1, t2), Bool
  | Call(f, el) -> check_call f el env 
  | Value(e1, e2) -> check_access_value e1 e2 env 
  | ListLit(el) -> check_list_lit el env  
  | DictLit(kvl) -> check_dict_lit kvl env
  | GraphLit(edgel) -> check_graph_lit edgel env
  | Null -> S_Null, Node
  | Noexpr -> S_Noexpr, Void 


(* Check list literal: all elements must be of the same type *)
and check_list_lit (el: Ast.expr list) (env: environment)  =
  if (List.length el) = 0 then
    S_ListLit([], Void), List(Void)
  else
    let (_, t1) = check_expr (List.hd el) env in
    let s_el = List.map (fun e -> 
      let (e2, t2) = check_expr e env in
        if t2 <> t1 then  
          raise (Error("Elements in a list must be" ^ " of the same type."));
      e2) el in
    S_ListLit(s_el, t1), List(t1)

(* Check if a type is boolean valued  *)
and check_bool_valued (t: Ast.var_type) =  match t with
    Bool | Int | Str
  | List(_) | Graph | Dict(_) | Node -> true
  | Float | Char | Void  -> false


and check_binop_type (t1: Ast.var_type) (op : Ast.op) (t2: Ast.var_type) (env:  
  environment) : (Ast.var_type) = 
  match op with 
    Plus -> 
      (try 
         type_of_math_expr t1 t2 
       with Not_found ->
        match (t1, t2) with 
           (Str, Str) -> Str
         | (Graph, Graph) -> Graph
         | (List(t1), List(t2)) when t1 = t2 -> List(t1)
         (* [|1, 2, 3 |] + [| |] empty list OK *)
         | (List(Void), List(typ)) | (List(typ), List(Void)) -> List(typ)
         | (Dict(t1, t2), Dict(t3, t4)) when t1 = t3 && t2 = t4 -> Dict(t1, t2)
         (* (|1:1|) + (| |) empty dict OK *)
         | (Dict(Void, Void), Dict(t1, t2)) | (Dict(t1, t2), Dict(Void, Void))
         -> Dict(t1, t2)
         | (_, _) -> raise (Error("Type mismatch in operator "
                            ^ Ast.string_of_binop op)))

  | Minus -> 
      (try 
         type_of_math_expr t1 t2 
       with Not_found ->
         if t1 <> Graph || t2 <> Graph then 
           raise (Error("Type mismatch in operator " ^ Ast.string_of_binop op));
         Graph) 

  | Multi -> type_of_math_expr t1 t2 
  | Divide -> type_of_math_expr t1 t2 
  | Mod -> type_of_math_expr t1 t2
  | Equal | Nequal -> 
       (try
         type_of_equal_expr t1 t2 
       with Not_found ->
         if t1 <> t2 then
           raise (Error("Type mismatch in operator " 
           ^ Ast.string_of_binop op));
         Bool)
  | Less | Leq | Greater | Geq ->
      (try
         type_of_equal_expr t1 t2 
       with Not_found ->
         match (t1, t2) with
           (Char, Char) -> Bool
         | (Str, Str) -> Bool
         | (_, _) -> raise (Error("Type mismatch in operator " 
                            ^ Ast.string_of_binop op)))
   | And | Or  -> 
      if not (check_bool_valued t1) || not (check_bool_valued t2) then
        raise (Error("Type mismatch in operator " ^ Ast.string_of_binop op));
      Bool


(* Check if t1 and t2 can be operands of arithmetic operations *)
and type_of_math_expr (t1: var_type) (t2: var_type) : (var_type) =
  match (t1, t2) with
    (Int, Int) -> Int
  | (Float, Float) -> Float 
  | (Int, Float) -> Float 
  | (Float, Int) -> Float 
  | (_, _) -> raise Not_found 



(* Check if t1 and t2 can be operands of equal *)
and type_of_equal_expr (t1: var_type) (t2: var_type) : (var_type) =
  match (t1, t2) with
    (Int, Int) -> Bool 
  | (Float, Float) -> Bool 
  | (Int, Float) -> Bool 
  | (Float, Int) -> Bool
  | (_, _) -> raise Not_found 


(* Check if t1 and t2 can be operands of in *)
and check_in_expr (t1: var_type) (t2: var_type) : bool =
  match (t1, t2) with
    (typ, List(vtyp)) when vtyp = typ -> true   (* element in list *)
  | (Str, Graph) -> true                        (* node in graph *)
  | (typ, Dict(kt, _)) when typ = kt -> true    (* key in dict *) 
  | (Str, Node) -> true                         (* attributes in node *)
  | (_, _) -> false


and  check_assign (e1: Ast.expr) (asnop: Ast.asnop) (e2: Ast.expr) (env:
  environment)  =
  let (e1, t1) = (check_expr e1 env) and (e2, t2) = (check_expr e2 env) in
  if assign_type_conversion t1 asnop t2 then  
    S_Assign(e1, asnop, e2, t1), t1
  else
    raise (Error("Cannot assign(" ^ string_of_asnop asnop ^ ") "
          ^  Ast.string_of_var_type t2 ^ " to type " 
    ^ Ast.string_of_var_type t1))


(* The callee must exist before the caller *)
and  check_call f (el: expr list) (env: environment) =
  let (el, el_t) = List.fold_right (fun e r -> 
    let (e, t) = check_expr e env in
    (e :: fst r, t :: snd r )) el ([], []) in 
  let ret_type = 
    try
      find_func f el_t env.funcs 
    with Not_found -> 
      let s = "Call( " ^ f ^ ", " ^ String.concat "; " (List.map
      string_of_var_type el_t) ^ ")" 
      in raise (Error("Function " ^ s  ^ " is no found"))
  in
  S_Call(f, el, ret_type), ret_type 

(* Check refernce: list[index], graph[node], dict[key], or node[attribute] *)
and check_access_value (e1: expr) (e2: expr) (env: environment) =
  let (e1, t1) = check_expr e1 env and (e2, t2) = check_expr e2 env in
  let typ = (match (t1, t2) with
               (List(vtyp), Int) -> vtyp
             | (Graph, Str) -> Node
             | (Dict(k1, vtyp), k2) when k1 = k2 -> vtyp 
             | (Node, Str) -> Str
             | (_, _) -> raise (Error("Type mismatch in reference"))) 
  in
  S_Value(e1, e2, typ, t1), typ

(* Check Dict literal: all keys / all values must be of the same types *)
and check_dict_lit (kvl: kv list) (env: environment) = 
  if (List.length kvl) = 0 then
    S_DictLit([], Void, Void), Dict(Void, Void)
  else
    let k_t = snd (check_expr (List.hd kvl).key env)
    and v_t = snd (check_expr (List.hd kvl).value env) in
    let el = List.map (
        fun kv -> 
          let (e1, t1) = check_expr kv.key env and (e2, t2) = check_expr
          kv.value env in
             if t1 <> k_t then raise (Error("Keys in a dict must be"  
                                      ^ " of the same type."));
             if t2 <> v_t then raise (Error("Values in a dict must be" 
                                      ^ " of the same type."));
          (e1, e2)) kvl 
    in 
    S_DictLit(el, k_t, v_t), Dict(k_t, v_t)

(* Check Graph literal: source/label/dest must be string valued *)
and check_graph_lit (edgel: edge list) (env: environment) =
  let edgel_t = List.map (fun e -> (check_expr e.src env, check_expr e.dest env,
  check_expr e.label env)) edgel in
  if (List.length (List.filter (fun (src, dest, label) -> snd src <> Str ||
  snd dest <> Str || snd label <> Str) edgel_t)) > 0 then
    raise (Error("Edges and nodes in a graph must be strings."))
  else
    S_GraphLit(List.map (fun ((e1,_), (e2,_), (e3,_)) -> (e1, e2, e3)) edgel_t)
    , Graph

(* Check return type *)
and check_return (e: Ast.expr) (env: environment) : s_stmt = 
  let (ret_e, ret_t) = check_expr e env in
  if ret_t = env.return_type then
    S_Return(ret_e) 
  else
    raise (Error("Cannot convert " ^ Ast.string_of_var_type ret_t ^ " to return type " 
    ^ Ast.string_of_var_type env.return_type))


(* Check for each loop: element in list, key in dict, node in graph *)
and check_foreach (e: Ast.expr) (s: Ast.stmt) (env: environment) =
  let Binop(e1, op, e2) = e in 
    if op <> In then
      raise (Error("Condition of a for-each loop must be (expr1 in expr2)")); 
    (match e1 with
      Id(_) -> 
          let scope1 = {parent = Some(env.scope); variables = [];} in
          let env1 = { env with scope = scope1; in_loop = true} in
          let (e1, t1) = check_expr e1 env1 and (e2, t2) = check_expr e2 env1 in
          (match t2 with 
            List(t) -> if t <> t1 then
                         raise (Error("Cannot convert type " ^ string_of_var_type  
                         t ^ " to type " ^ string_of_var_type t1 ));
                       S_Foreach(e1, e2, t2, check_stmt s env1)
          | Dict(t, _) -> 
              if t <> t1 then
                 raise (Error("Cannot convert type " ^ string_of_var_type t
                 ^ " to type " ^ string_of_var_type t1 ));
              S_Foreach(e1, e2, t2, check_stmt s env1) 
          | Graph -> S_Foreach(e1, e2, t2, check_stmt s env1) 

          | _ -> raise (Error("For-each loop must iterate a list/dict/graph")))
     | _ -> raise (Error("First expression of a for-each loop must be an identifier"))) 


(* Check value declaration *)
and check_var_decl (v: Ast.var_decl) (env: environment) : (s_var_decl) =  
  let exist = List.exists (fun (name, _) -> name = v.v_name) env.scope.variables
  in
  if exist then
    raise (Error("Indentifier already declared"))
  else
    env.scope.variables <- (v.v_name, v.v_type) :: env.scope.variables;
    let (s_vinit, t) = check_expr v.v_init env in 
      if assign_type_conversion v.v_type Asn t then  
        {s_vtype = v.v_type; s_vname = v.v_name;
         s_vinit = s_vinit; }
      else 
        raise (Error("Variable initialization: cannot convert " 
        ^ Ast.string_of_var_type t 
        ^ " to type " ^ Ast.string_of_var_type v.v_type))  


(* Check if t2 type can be assigned to t1 type *)
and assign_type_conversion (t1: Ast.var_type) (asnop: Ast.asnop) (t2: Ast.var_type) : bool =
  match asnop with
    Asn ->
      (if t1 = t2 then
        true
      else
        match (t1, t2) with
          (_, Void) -> true
        | (List(_), List(Void)) -> true             (* empty list *)
        | (Dict(_, _), Dict(Void, Void)) -> true    (* empty dict *)
        | (Float, Int) -> true                      (* float a = 3 OK *)
        | _ -> false) 
  | Add ->
      (if t1 = Char || t1 = Bool then
        false
      else if t1 = t2 then
        true
      else
        match (t1, t2) with
          (List(_), List(Void)) -> true             (* empty list *)
        | (Dict(_, _), Dict(Void, Void)) -> true    (* empty dict *)
        | (Float, Int) -> true                      (* float a += 3*)
        | _ -> false) 
 | Sub -> 
        (match (t1, t2) with
          (Int, Int) -> true
        | (Float, Float) -> true
        | (Float, Int) -> true
        | (Graph, Graph) -> true                    (* graph -= graph *)
        | _ -> false)

(* Check statements *) 
and check_stmt (s: Ast.stmt) (env: environment) = match s with
    Block(slist) ->
      let scope1 = {parent = Some(env.scope); variables = [];} in
      let env1 = { env with scope = scope1} in
      let slist = List.map (fun s -> check_stmt s env1) slist in
      scope1.variables <- List.rev scope1.variables;
      S_Block(scope1, slist)  
  | Variable(v) -> S_Variable(check_var_decl v env) 
  | Expr(e) -> let (e1, t1) = check_expr e env in S_Expr(e1, t1) 
  | Return(e) -> check_return e env
  | If(e, s1, s2) -> 
      let (e, t) = (check_expr e env) in
        if not (check_bool_valued t) then
          raise (Error("Condition must be boolean-valued"));
      S_If(e, check_stmt s1 env, check_stmt s2 env) 
  | For(e1, e2, e3, s) -> 
      let scope1 = {parent = Some(env.scope); variables = [];} in
      let env1 = { env with scope = scope1; in_loop = true} in
      let (e2, t) = check_expr e2 env1 in
        if not (check_bool_valued t) then
          raise (Error("Condition must be boolean-valued"));
      S_For(fst (check_expr e1 env), e2, fst (check_expr e3 env), check_stmt s env)
  | Foreach(e, s) -> check_foreach e s env
  | While(e, s) -> 
      let scope1 = {parent = Some(env.scope); variables = [];} in
      let env1 = { env with scope = scope1; in_loop = true} in
      let (e, t) = (check_expr e env1) in
        if not (check_bool_valued t) then
          raise (Error("Condition must be boolean-valued"));
      S_While(e, check_stmt s env1) 
  | Continue -> 
      if env.in_loop then
        S_Continue
      else
        raise (Error("Continue statement must be in loop"))
  | Break ->
      if env.in_loop then
        S_Break
      else
        raise (Error("Break statement must be in loop"))
  
(* Check function declaraion *) 
let check_func_decl (func: Ast.func_decl) (env: environment) : (s_func_decl) =
  let scope1 = {parent = Some(env.scope); variables = [];} in
  let env1 = {env with scope = scope1; return_type = func.rtype }  in 
  let fname = (if func.fname = "main" then "__main" else func.fname) in 
  let formals_t = List.map (fun v -> (check_var_decl v env1).s_vtype) func.formals
  in
  try 
    ignore(find_func fname formals_t env.funcs);
    raise (Error("Duplicate function definition for " ^ fname)); 
  with Not_found -> 
    let fdecl = {
      s_fname = fname; 
      s_formals = (List.map (fun var -> (var.v_type, var.v_name)) func.formals);
      s_body = (List.map (fun stmt -> check_stmt stmt env1) func.body);
      s_rtype = func.rtype; 
    } in
    env.funcs <- (fdecl.s_rtype, fdecl.s_fname, formals_t) ::env.funcs; fdecl
  | _ -> raise (Error("Duplicate function definition for " ^ fname)) 

(* Initial environment *)
let init_env : (environment) = 
  let std_funcs = [
                   (* Print *)
                   (Void, "print", [Str]);
                   (Void, "print", [Int]);
                   (Void, "print", [Float]);
                   (Void, "print", [Bool]);
                   (Void, "print", [Char]);
                   (Void, "print", [Graph]);
                   (Void, "print", [Node]);

                   (* Node  *)
                   (Str, "getName", [Node]);
                   (List(Node), "getOutNeighbors", [Node; Str]);
                   (List(Node), "getOutNeighbors", [Node]);
                   (List(Node), "getInNeighbors", [Node; Str]);
                   (List(Node), "getInNeighbors", [Node]);
                   (Dict(Str, Str), "getAttributes", [Node]);

                   (* Graph *)
                   (List(Str), "getLabels", [Graph]);
                   (List(Node), "getNodes", [Graph]);

                   (* List *)
                   (Int, "getSize", [List(Int)]);
                   (Int, "getSize", [List(Float)]);
                   (Int, "getSize", [List(Char)]);
                   (Int, "getSize", [List(Str)]);
                   (Int, "getSize", [List(Bool)]);
                   (Int, "getSize", [List(Node)]);
                 ] in 
                    
  let init_scope = { parent = None; variables = []; } in
  {
    scope = init_scope;
    funcs = std_funcs;
    return_type = Void;
    in_loop = false;
  }

let check_program (program: Ast.program) : (Sast.s_program) =
  let env = init_env in
  let (vdecls, fdecls) = program in 
  {
    s_var_decls = List.map (fun var -> check_var_decl var env) (List.rev vdecls); 
    s_func_decls = List.map (fun func -> check_func_decl func env) (List.rev
    fdecls);
    s_symbols = env.scope
  } 
